import gdsfactory as gf
from gdsfactory.gdsdiff.gdsdiff import gdsdiff


def test_differences():
    c1 = gf.components.straight(length=2)
    c2 = gf.components.straight(length=3)
    c = gdsdiff(c1, c2)
    assert c.references[-1].area() == 0.5


def test_no_differences():
    c1 = gf.components.straight(length=2)
    c2 = gf.components.straight(length=2)
    c = gdsdiff(c1, c2)
    assert c.references[-1].area() == 0


if __name__ == "__main__":
    # test_no_differences()
    # test_differences()
    c1 = gf.components.straight(length=2)
    c2 = gf.components.straight(length=2)
    c = gdsdiff(c1, c2)
    c.show()



"""Sweep neff over wavelength and returns group index."""

from functools import partial

from gdsfactory.simulation.gmeep.get_material import get_index
from gdsfactory.simulation.modes.find_modes import find_modes
from gdsfactory.simulation.modes.types import Mode


def find_mode_dispersion(
    wavelength: float = 1.55,
    wavelength_step: float = 0.01,
    core: str = "Si",
    clad: str = "SiO2",
    mode_number: int = 1,
    **kwargs,
) -> Mode:
    """Returns Mode with correct dispersion (ng)

    group index comes from a finite difference approximation at 3 wavelengths

    Args:
        wavelength: center wavelength (um)
        wavelength_step: in um
        core: core material name
        clad: clad material name
        mode_number: mode index to compute (1: fundanmental mode)

    Keyword Args:
        wg_thickness: wg height (um)
        sx: supercell width (um)
        sy: supercell height (um)
        resolution: (pixels/um)
        wavelength: wavelength
        num_bands: mode order
        plot: if True plots mode
        logscale: plots in logscale
        plotH: plot magnetic field
        dirpath: path to save the modes
        polarization: prefix when saving the modes
        paririty: symmetries mp.ODD_Y mp.EVEN_X for TE, mp.EVEN_Y for TM

    """
    w0 = wavelength - wavelength_step
    wc = wavelength
    w1 = wavelength + wavelength_step

    ncore = partial(get_index, name=core)
    nclad = partial(get_index, name=clad)

    m0 = find_modes(wavelength=w0, ncore=ncore(w0), nclad=nclad(w0), **kwargs)
    mc = find_modes(wavelength=wc, ncore=ncore(wc), nclad=nclad(wc), **kwargs)
    m1 = find_modes(wavelength=w1, ncore=ncore(w1), nclad=nclad(w1), **kwargs)

    n0 = m0[mode_number].neff
    nc = mc[mode_number].neff
    n1 = m1[mode_number].neff

    # ng = ncenter - wavelength *dn/ step
    ng = nc - wavelength * (n1 - n0) / (2 * wavelength_step)
    neff = (n0 + nc + n1) / 3
    return Mode(mode_number=mode_number, ng=ng, neff=neff, wavelength=wavelength)


if __name__ == "__main__":
    m = find_mode_dispersion(wg_width=0.45, wg_thickness=0.22)
    print(m.ng)
    # test_ng()
    # print(get_index(name="Si"))
    # ngs = []
    # for wavelength_step in [0.001, 0.01]:
    #     neff, ng = find_modes_dispersion(
    #         wg_width=0.45, wg_thickness=0.22, wavelength_step=wavelength_step
    #     )
    #     ngs.append(ng)
    #     print(wavelength_step, ng)



from typing import Dict, List, Union

import gdsfactory as gf
from gdsfactory.components.straight import straight
from gdsfactory.difftest import difftest
from gdsfactory.port import Port
from gdsfactory.types import ComponentOrFactory, Routes


def get_routes_straight(
    ports: Union[List[Port], Dict[str, Port]],
    straight: ComponentOrFactory = straight,
    **kwargs,
) -> Routes:
    """Returns routes made by 180 degree straights.

    Args:
        ports: List or dict of ports
        straight: function for straight
        **kwargs: waveguide settings
    """
    ports = list(ports.values()) if isinstance(ports, dict) else ports
    straight = straight(**kwargs)
    references = [straight.ref() for port in ports]
    references = [ref.connect("o1", port) for port, ref in zip(ports, references)]
    ports = [ref.ports["o2"] for i, ref in enumerate(references)]
    lengths = [straight.info.length] * len(ports)
    return Routes(references=references, ports=ports, lengths=lengths)


def test_get_routes_straight(check: bool = True):
    c = gf.Component("get_routes_straight")
    pad_array = gf.components.pad_array()
    c1 = c << pad_array
    c2 = c << pad_array
    c2.ymax = -200

    routes = get_routes_straight(ports=c1.get_ports_list(), length=200)
    c.add(routes.references)
    if check:
        difftest(c)
    return c


if __name__ == "__main__":
    c = test_get_routes_straight(False)
    c.show()



import toolz

import gdsfactory as gf

extend_ports1 = gf.partial(gf.components.extend_ports, length=1)
extend_ports2 = gf.partial(gf.components.extend_ports, length=10)


straigth_extended1 = toolz.compose(
    extend_ports1, gf.partial(gf.components.straight, width=0.5)
)
straigth_extended2 = toolz.compose(
    extend_ports2, gf.partial(gf.components.straight, width=0.9)
)

straigth_extended3 = toolz.compose(
    extend_ports2, gf.partial(gf.components.straight, width=0.5)
)


def test_compose1():
    """Ensures the first level of composed function gets a unique name"""
    mzi500 = gf.partial(gf.components.mzi, straight=straigth_extended1)
    mzi900 = gf.partial(gf.components.mzi, straight=straigth_extended2)

    c500 = mzi500()
    c900 = mzi900()

    assert c900.name != c500.name, f"{c500.name} must be different from {c900.name}"


def test_compose2():
    """Ensures the second level of composed function gets a unique name.

    FIXME! this one does not work

    """
    mzi500 = gf.partial(gf.components.mzi, straight=straigth_extended3)
    mzi900 = gf.partial(gf.components.mzi, straight=straigth_extended2)

    c500 = mzi500()
    c900 = mzi900()

    assert c900.name != c500.name, f"{c500.name} must be different from {c900.name}"


if __name__ == "__main__":
    mzi500 = gf.partial(gf.components.mzi, straight=straigth_extended3)
    mzi900 = gf.partial(gf.components.mzi, straight=straigth_extended2)

    c900 = mzi900()
    c500 = mzi500()

    c = gf.Component()
    r500 = c << c500
    r900 = c << c900
    r900.ymin = r500.ymax + 10
    c.show()

    assert c900.name != c500.name, f"{c500.name} must be different from {c900.name}"



import re

import numpy as np
import pandas as pd
import tidy3d as td
from tqdm import tqdm

from gdsfactory.simulation.gtidy3d.get_results import get_results
from gdsfactory.simulation.gtidy3d.get_simulation import get_simulation
from gdsfactory.types import Component, Optional, PortSymmetries


def parse_port_eigenmode_coeff(port_index: int, ports, sim_data: td.SimulationData):
    """Given a port and eigenmode coefficient result, returns the coefficients
    relative to whether the wavevector is entering or exiting simulation

    Args:
        port_index: index of port
        ports: component_ref.ports
        sim_data:
    """
    if f"o{port_index}" not in ports:
        raise ValueError(
            f"port = 'o{port_index}' not in {list(ports.keys())}. "
            "You can rename ports with Component.auto_rename_ports()"
        )

    # Direction of port (pointing away from the simulation)
    # Figure out if that is exiting the simulation or not
    # depending on the port orientation (assuming it's near PMLs)
    if ports[f"o{port_index}"].orientation == 0:  # east
        direction_inp = "-"
        direction_out = "+"
    elif ports[f"o{port_index}"].orientation == 90:  # north
        direction_inp = "-"
        direction_out = "+"
    elif ports[f"o{port_index}"].orientation == 180:  # west
        direction_inp = "+"
        direction_out = "-"
    elif ports[f"o{port_index}"].orientation == 270:  # south
        direction_inp = "+"
        direction_out = "-"
    else:
        ValueError(
            "Port orientation {ports[port_index].orientation} is not 0, 90, 180, or 270 degrees!"
        )

    coeff_inp = sim_data.monitor_data[f"o{port_index}"].amps.sel(
        direction=direction_inp
    )
    coeff_out = sim_data.monitor_data[f"o{port_index}"].amps.sel(
        direction=direction_out
    )
    return coeff_inp, coeff_out


def get_sparameters(
    component: Component, port_symmetries: Optional[PortSymmetries] = None, **kwargs
) -> pd.DataFrame:
    """
    Get full sparameter matrix
    Simulates each time using a different input port (by default, all of them)
    unless you specify port_symmetries:

    port_symmetries = {"o1":
            {
                "s11": ["s22","s33","s44"],
                "s21": ["s21","s34","s43"],
                "s31": ["s13","s24","s42"],
                "s41": ["s14","s23","s32"],
            }
        }
    - Only simulations using the outer key port names will be run
    - The associated value is another dict whose keys are the S-parameters computed
        when this source is active
    - The values of this inner Dict are lists of s-parameters whose values are copied

    Args:
        component: to simulate.
        port_symmetries: Dict to specify port symmetries, to save number of simulations

    Keyword Args:
        mode_index: mode index
        n_modes: number of modes
        port_extension: extend ports beyond the PML
        layer_stack: contains layer numbers (int, int) to thickness, zmin
        zmargin: thickness for cladding above and below core
        thickness_pml: PML thickness (um)
        clad_material: material for cladding
        port_source_name: input port name
        port_margin: margin on each side of the port
        distance_source_to_monitors: in (um) source goes before monitors
        resolution: grid_size=3*[1/resolution]
        wavelength: in (um)
        plot_modes: plot source modes.

    """

    port_symmetries = port_symmetries or {}
    monitor_indices = []
    source_indices = []
    component_ref = component.ref()

    for port_name in component_ref.ports.keys():
        if component_ref.ports[port_name].port_type == "optical":
            monitor_indices.append(re.findall("[0-9]+", port_name)[0])
    if bool(port_symmetries):  # user-specified
        for port_name in port_symmetries.keys():
            source_indices.append(re.findall("[0-9]+", port_name)[0])
    else:  # otherwise cycle through all
        source_indices = monitor_indices

    num_sims = len(port_symmetries.keys()) or len(source_indices)
    sp = {}

    def get_sparameter(
        n: int,
        component: Component,
        port_symmetries=port_symmetries,
        monitor_indices=monitor_indices,
        **kwargs,
    ) -> np.ndarray:
        """Return Component sparameter for a particular port Index n

        Args:
            n: port_index
            component:
            port_symmetries:
            monitor_indices:

        """
        sim = get_simulation(
            component, port_source_name=f"o{monitor_indices[n]}", **kwargs
        )
        sim_data = get_results(sim).result()
        source_entering, source_exiting = parse_port_eigenmode_coeff(
            monitor_indices[n], component_ref.ports, sim_data
        )

        for monitor_index in monitor_indices:
            j = monitor_indices[n]
            i = monitor_index
            if monitor_index == monitor_indices[n]:
                sii = source_exiting / source_entering

                siia = np.unwrap(np.angle(sii))
                siim = np.abs(sii)

                sp[f"s{i}{i}a"] = siia
                sp[f"s{i}{i}m"] = siim
            else:
                monitor_entering, monitor_exiting = parse_port_eigenmode_coeff(
                    monitor_index, component_ref.ports, sim_data
                )
                sij = monitor_exiting / source_entering
                sija = np.unwrap(np.angle(sij))
                sijm = np.abs(sij)
                sp[f"s{i}{j}a"] = sija
                sp[f"s{i}{j}m"] = sijm
                sij = monitor_entering / source_entering
                sija = np.unwrap(np.angle(sij))
                sijm = np.abs(sij)

        if bool(port_symmetries) is True:
            for key in port_symmetries[f"o{monitor_indices[n]}"].keys():
                values = port_symmetries[f"o{monitor_indices[n]}"][key]
                for value in values:
                    sp[f"{value}m"] = sp[f"{key}m"]
                    sp[f"{value}a"] = sp[f"{key}a"]

        return sp

    for n in tqdm(range(num_sims)):
        sp.update(
            get_sparameter(
                n,
                component=component,
                port_symmetries=port_symmetries,
                monitor_indices=monitor_indices,
                **kwargs,
            )
        )
    return sp


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.components.straight(length=2)
    s = get_sparameters(c)
    print(s)



"""
find GDS labels and write labels into a CSV file
"""

import csv
import pathlib
from pathlib import Path
from typing import Iterator, Tuple

from loguru import logger

from gdsfactory import LAYER
from gdsfactory.routing.add_fiber_single import add_fiber_single
from gdsfactory.types import Optional, PathType


def find_labels(
    gdspath: Path, layer_label: Tuple[int, int] = LAYER.LABEL, prefix: str = "opt_"
) -> Iterator[Tuple[str, float, float]]:
    """finds labels and locations from a GDS file"""
    import klayout.db as pya

    # Load the layout
    gdspath = str(gdspath)
    layout = pya.Layout()
    layout.read(gdspath)

    # Get the top cell and the units, and find out the index of the layer
    topcell = layout.top_cell()
    dbu = layout.dbu
    layer = pya.LayerInfo(layer_label[0], layer_label[1])
    layer_index = layout.layer(layer)

    # Extract locations
    iterator = topcell.begin_shapes_rec(layer_index)

    while not (iterator.at_end()):
        shape, trans = iterator.shape(), iterator.trans()
        iterator.next()
        if shape.is_text():
            text = shape.text
            if text.string.startswith(prefix):
                transformed = text.transformed(trans)
                yield text.string, transformed.x * dbu, transformed.y * dbu


def write_labels(
    gdspath: Path,
    layer_label: Tuple[int, int] = LAYER.TEXT,
    filepath: Optional[PathType] = None,
    prefix: str = "opt_",
) -> Path:
    """Load  GDS mask and extracts the labels and coordinates from a GDS file"""
    labels = list(find_labels(gdspath, layer_label=layer_label, prefix=prefix))
    gdspath = pathlib.Path(gdspath)

    filepath = filepath or gdspath.with_suffix(".csv")
    with open(filepath, "w", newline="") as f:
        writer = csv.writer(f)
        writer.writerows(labels)
    logger.info(f"Wrote {len(labels)} labels in {filepath.absolute()}")
    return filepath


def test_find_labels():
    import gdsfactory as gf

    c = gf.components.straight(length=124)
    cc = add_fiber_single(component=c)
    gdspath = cc.write_gds()
    assert len(list(find_labels(gdspath))) == 4


if __name__ == "__main__":
    test_find_labels()

    # import gdsfactory as gf

    # c = gf.components.straight()
    # cc = add_fiber_single(component=c)
    # gdspath = cc.write_gds()
    # print(len(list(find_labels(gdspath))))
    # cc.show()

    # gdspath = CONFIG["samples_path"] / "mask" / "build" / "mask" / "sample.gds"
    # write_labels(gdspath)



"""Lets add pins to each cell from the fab a PDK.

"""

import gdsfactory as gf
from gdsfactory.add_pins import add_outline, add_pins
from gdsfactory.cross_section import strip
from gdsfactory.difftest import difftest

WIDTH = 2
LAYER = (30, 0)

fab_a_metal = gf.partial(strip, width=WIDTH, layer=LAYER)
fab_a_metal.__name__ = "fab_a_metal"


def test_waveguide():
    c = gf.components.straight(cross_section=fab_a_metal)
    difftest(c)


def decorator(component) -> None:
    """Fab specific functions over a component."""
    add_pins(component)
    add_outline(component)


mmi2x2 = gf.partial(gf.components.mmi2x2, decorator=decorator)
mmi1x2 = gf.partial(gf.components.mmi1x2, decorator=decorator)
bend_euler = gf.partial(gf.components.bend_euler, decorator=decorator)
straight = gf.partial(gf.components.straight, decorator=decorator)
mzi = gf.partial(gf.components.mzi, splitter=mmi1x2, bend=bend_euler, straight=straight)

factory = dict(mmi2x2=mmi2x2, mmi1x2=mmi1x2, mzi=mzi)


if __name__ == "__main__":
    c = mzi()
    c.show()



"""You can define a path with a list of points combined with a cross-section.

A path can be extruded using any CrossSection returning a Component

The CrossSection defines the layer numbers, widths and offsetts

Based on phidl.path
"""

from collections.abc import Iterable
from typing import Optional

import numpy as np
import phidl.path as path
from phidl.device_layout import Path as PathPhidl
from phidl.device_layout import _simplify
from phidl.path import smooth as smooth_phidl

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.cross_section import CrossSection
from gdsfactory.types import (
    Coordinates,
    CrossSectionOrFactory,
    Float2,
    Layer,
    PathFactory,
)


class Path(PathPhidl):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        """pydantic assumes Path is always valid"""
        return v


def _sinusoidal_transition(y1, y2):
    dy = y2 - y1

    def sine(t):
        return y1 + (1 - np.cos(np.pi * t)) / 2 * dy

    return sine


def _linear_transition(y1, y2):
    dy = y2 - y1

    def linear(t):
        return y1 + t * dy

    return linear


def transition_exponential(y1, y2, exp=0.5):
    """Returns the function for an exponential transition

    Args:
        y1: start width
        y2: end width
        exp: exponent
    """
    return lambda t: y1 + (y2 - y1) * t ** exp


def transition(
    cross_section1: CrossSection,
    cross_section2: CrossSection,
    width_type: str = "sine",
) -> CrossSection:
    """Creates a CrossSection that smoothly transitions between two input
    CrossSections. Only cross-sectional elements that have the `name` (as in
    X.add(..., name = 'wg') ) parameter specified in both input CrosSections
    will be created. Port names will be cloned from the input CrossSections in
    reverse.
    adapted from phidl.path

    Args:
        cross_section1: First input CrossSection
        cross_section2: Second input CrossSection
        width_type: sine or linear
          Sets the type of width transition used if any widths are different
          between the two input CrossSections.

    Returns A smoothly-transitioning CrossSection
    """

    X1 = cross_section1
    X2 = cross_section2
    Xtrans = CrossSection()

    if not X1.aliases or not X2.aliases:
        raise ValueError(
            """transition() found no named sections in one
        or both inputs (cross_section1/cross_section2)."""
        )

    layers1 = {section["layer"] for section in X1.sections}
    layers2 = {section["layer"] for section in X2.sections}

    has_common_layers = True if layers1.intersection(layers2) else False
    if not has_common_layers:
        raise ValueError(
            f"transition() found no common layers X1 {layers1} and X2 {layers2}"
        )

    for alias in X1.aliases.keys():
        if alias in X2.aliases:

            offset1 = X1[alias]["offset"]
            offset2 = X2[alias]["offset"]
            width1 = X1[alias]["width"]
            width2 = X2[alias]["width"]

            if callable(offset1):
                offset1 = offset1(1)
            if callable(offset2):
                offset2 = offset2(0)
            if callable(width1):
                width1 = width1(1)
            if callable(width2):
                width2 = width2(0)

            offset_fun = _sinusoidal_transition(offset1, offset2)

            if width_type == "sine":
                width_fun = _sinusoidal_transition(width1, width2)
            elif width_type == "linear":
                width_fun = _linear_transition(width1, width2)
            else:
                raise ValueError(
                    "transition() width_type "
                    "argument must be one of {'sine','linear'}"
                )

            if X1[alias]["layer"] != X2[alias]["layer"]:
                hidden = True
                layer = (X1[alias]["layer"], X2[alias]["layer"])
            else:
                hidden = False
                layer = X1[alias]["layer"]

            Xtrans.add(
                width=width_fun,
                offset=offset_fun,
                layer=layer,
                ports=(X2[alias]["ports"][0], X1[alias]["ports"][1]),
                port_types=(X2[alias]["port_types"][0], X1[alias]["port_types"][1]),
                name=alias,
                hidden=hidden,
            )

    Xtrans.cross_sections = (X1, X2)
    Xtrans.name = f"trans_{width_type}_{X1.get_name()}_{X2.get_name()}"
    return Xtrans


@cell
def extrude(
    p: Path,
    cross_section: Optional[CrossSectionOrFactory] = None,
    layer: Optional[Layer] = None,
    width: Optional[float] = None,
    widths: Optional[Float2] = None,
    simplify: Optional[float] = None,
) -> Component:
    """Returns Component extruding a Path with a cross_section.

    A path can be extruded using any CrossSection returning a Component

    The CrossSection defines the layer numbers, widths and offsetts

    adapted from phidl.path

    Args:
        p: a path is a list of points (arc, straight, euler)
        cross_section: to extrude
        layer:
        width:
        widths: tuple of starting and end width
        simplify: Tolerance value for the simplification algorithm.
          All points that can be removed without changing the resulting
          polygon by more than the value listed here will be removed.
    """
    if cross_section is None and layer is None:
        raise ValueError("CrossSection or layer needed")

    if cross_section is not None and layer is not None:
        raise ValueError("Define only CrossSection or layer")

    if layer is not None and width is None and widths is None:
        raise ValueError("Need to define layer width or widths")
    elif width:
        cross_section = CrossSection()
        cross_section.add(width=width, layer=layer)

    elif widths:
        cross_section = CrossSection()
        cross_section.add(width=_linear_transition(widths[0], widths[1]), layer=layer)

    xsection_points = []
    c = Component()

    cross_section = cross_section() if callable(cross_section) else cross_section
    snap_to_grid = cross_section.info.get("snap_to_grid", None)

    for section in cross_section.sections:
        width = section["width"]
        offset = section["offset"]
        layer = section["layer"]
        ports = section["ports"]
        port_types = section["port_types"]
        hidden = section["hidden"]

        if isinstance(width, (int, float)) and isinstance(offset, (int, float)):
            xsection_points.append([width, offset])
        if isinstance(layer, int):
            layer = (layer, 0)
        if (
            isinstance(layer, Iterable)
            and len(layer) == 2
            and isinstance(layer[0], int)
            and isinstance(layer[1], int)
        ):
            xsection_points.append([layer[0], layer[1]])

        # print(offset, type(offset))
        if callable(offset):
            P_offset = p.copy()
            P_offset.offset(offset)
            points = P_offset.points
            start_angle = P_offset.start_angle
            end_angle = P_offset.end_angle
            offset = 0
        else:
            points = p.points
            start_angle = p.start_angle
            end_angle = p.end_angle

        if callable(width):
            # Compute lengths
            dx = np.diff(p.points[:, 0])
            dy = np.diff(p.points[:, 1])
            lengths = np.cumsum(np.sqrt(dx ** 2 + dy ** 2))
            lengths = np.concatenate([[0], lengths])
            width = width(lengths / lengths[-1])
        else:
            pass

        points1 = p._centerpoint_offset_curve(
            points,
            offset_distance=offset + width / 2,
            start_angle=start_angle,
            end_angle=end_angle,
        )
        points2 = p._centerpoint_offset_curve(
            points,
            offset_distance=offset - width / 2,
            start_angle=start_angle,
            end_angle=end_angle,
        )

        # Simplify lines using the Ramer–Douglas–Peucker algorithm
        if isinstance(simplify, bool):
            raise ValueError(
                "[PHIDL] the simplify argument must be a number (e.g. 1e-3) or None"
            )
        if simplify is not None:
            points1 = _simplify(points1, tolerance=simplify)
            points2 = _simplify(points2, tolerance=simplify)

        if snap_to_grid:
            snap_to_grid_nm = snap_to_grid * 1e3
            points1 = (
                snap_to_grid_nm
                * np.round(np.array(points1) * 1e3 / snap_to_grid_nm)
                / 1e3
            )
            points2 = (
                snap_to_grid_nm
                * np.round(np.array(points2) * 1e3 / snap_to_grid_nm)
                / 1e3
            )

        # Join points together
        points = np.concatenate([points1, points2[::-1, :]])

        layers = layer if hidden else [layer, layer]
        if not hidden and p.length() > 1e-3:
            c.add_polygon(points, layer=layer)

        # Add ports if they were specified
        if ports[0] is not None:
            orientation = (p.start_angle + 180) % 360
            _width = width if np.isscalar(width) else width[0]
            new_port = c.add_port(
                name=ports[0],
                layer=layers[0],
                port_type=port_types[0],
                width=_width,
                orientation=orientation,
                cross_section=cross_section.cross_sections[0]
                if hasattr(cross_section, "cross_sections")
                else cross_section,
            )
            new_port.endpoints = (points1[0], points2[0])
        if ports[1] is not None:
            orientation = (p.end_angle + 180) % 360
            _width = width if np.isscalar(width) else width[-1]
            new_port = c.add_port(
                name=ports[1],
                layer=layers[1],
                port_type=port_types[1],
                width=_width,
                orientation=orientation,
                cross_section=cross_section.cross_sections[1]
                if hasattr(cross_section, "cross_sections")
                else cross_section,
            )
            new_port.endpoints = (points2[-1], points1[-1])

    # points = np.concatenate((p.points, np.array(xsection_points)))
    # points_hash = hash_points(points)[:26]
    # name = f"path_{points_hash}"
    # c.info.points_hash = points_hash
    # clean_dict(p.info)
    # clean_dict(cross_section.info)
    # c.info.path = p.info
    # c.info.cross_section = cross_section.info

    c.info.length = float(np.round(p.length(), 3))
    return c


def arc(radius: float = 10.0, angle: int = 90, npoints: int = 720) -> Path:
    """Returns a radial arc.

    Args:
        radius: minimum radius of curvature
        angle: total angle of the curve
        npoints: Number of points used per 360 degrees

    """
    return path.arc(radius=radius, angle=angle, num_pts=npoints)


def euler(
    radius: float = 10,
    angle: int = 90,
    p: float = 0.5,
    use_eff: bool = False,
    npoints: int = 720,
) -> Path:
    """Returns an euler bend that adiabatically transitions from straight to curved.
    By default, `radius` corresponds to the minimum radius of curvature of the bend.
    However, if `use_eff` is set to True, `radius` corresponds to the effective
    radius of curvature (making the curve a drop-in replacement for an arc). If
    p < 1.0, will create a "partial euler" curve as described in Vogelbacher et.
    al. https://dx.doi.org/10.1364/oe.27.031394

    Args:
        radius: minimum radius of curvature
        angle: total angle of the curve
        p: Proportion of the curve that is an Euler curve
        use_eff: If False: `radius` is the minimum radius of curvature of the bend
            If True: The curve will be scaled such that the endpoints match an arc
            with parameters `radius` and `angle`
        npoints: Number of points used per 360 degrees

    """
    return path.euler(radius=radius, angle=angle, p=p, use_eff=use_eff, num_pts=npoints)


def straight(length: float = 10.0, npoints: int = 2) -> Path:
    """Returns a straight path

    For transitions you should increase have at least 100 points

    Args:
        length: of straight
        npoints: number of points
    """
    if length < 0:
        raise ValueError(f"length = {length} needs to be > 0")
    return path.straight(length=length, num_pts=npoints)


def smooth(
    points: Coordinates,
    radius: float = 4.0,
    bend: PathFactory = euler,
    **kwargs,
) -> Path:
    """Returns a smooth Path from a series of waypoints. Corners will be rounded
    using `bend` and any additional key word arguments (for example,
    `use_eff = True` for `bend = gf.path.euler`)

    Args:
        points: array-like[N][2] List of waypoints for the path to follow
        radius: radius of curvature, passed to `bend`
        bend: bend function to round corners
        **kwargs: Extra keyword arguments that will be passed to `bend`
    """
    return smooth_phidl(points=points, radius=radius, corner_fun=bend, **kwargs)


__all__ = ["straight", "euler", "arc", "extrude", "path", "transition", "smooth"]


def _demo():
    import gdsfactory as gf

    c = gf.Component()
    X1 = gf.CrossSection()
    X1.add(width=1.2, offset=0, layer=2, name="wg", ports=("in1", "out1"))
    X1.add(width=2.2, offset=0, layer=3, name="etch")
    X1.add(width=1.1, offset=3, layer=1, name="wg2")

    # Create the second CrossSection that we want to transition to
    X2 = gf.CrossSection()
    X2.add(width=1, offset=0, layer=2, name="wg", ports=("in2", "out2"))
    X2.add(width=3.5, offset=0, layer=3, name="etch")
    X2.add(width=3, offset=5, layer=1, name="wg2")

    Xtrans = gf.path.transition(cross_section1=X1, cross_section2=X2, width_type="sine")

    P1 = gf.path.straight(length=5)
    P2 = gf.path.straight(length=5)

    wg1 = gf.path.extrude(P1, X1)
    wg2 = gf.path.extrude(P2, X2)

    P4 = gf.path.euler(radius=25, angle=45, p=0.5, use_eff=False)
    wg_trans = gf.path.extrude(P4, Xtrans)
    wg1_ref = c << wg1
    wg2_ref = c << wg2
    wgt_ref = c << wg_trans
    wgt_ref.connect("in2", wg1_ref.ports["out1"])
    wg2_ref.connect("in2", wgt_ref.ports["out1"])

    print(wg1)
    print(wg2)
    print(wg_trans)
    c.show()


if __name__ == "__main__":

    # P = euler(radius=10, use_eff=True)
    # P = euler()
    # P = Path()
    # P.append(straight(length=5))
    # P.append(path.arc(radius=10, angle=90))
    # P.append(path.spiral())

    # Create a blank CrossSection
    # X = CrossSection()
    # X.add(width=0.5, offset=0, layer=LAYER.SLAB90, ports=["in", "out"])

    # X.add(width=2.0, offset=-4, layer=LAYER.HEATER, ports=["HW1", "HE1"])
    # X.add(width=2.0, offset=4, layer=LAYER.HEATER, ports=["HW0", "HE0"])

    # Combine the Path and the CrossSection

    # c = extrude(P, X, simplify=5e-3)
    # c = gf.add_pins(c)
    # c << gf.components.bend_euler(radius=10)
    # c << gf.components.bend_circular(radius=10)
    # print(c.ports["in"].layer)
    # c.show()

    import gdsfactory as gf

    X1 = gf.CrossSection()
    X1.add(width=1.2, offset=0, layer=2, name="wg", ports=("o1", "o2"))
    X1.add(width=2.2, offset=0, layer=3, name="etch")
    X1.add(width=1.1, offset=3, layer=1, name="wg2")

    X2 = gf.CrossSection()
    X2.add(width=1, offset=0, layer=2, name="wg", ports=("o1", "o2"))
    X2.add(width=3.5, offset=0, layer=3, name="etch")
    X2.add(width=3, offset=5, layer=1, name="wg2")
    Xtrans = gf.path.transition(cross_section1=X1, cross_section2=X2, width_type="sine")

    P1 = gf.path.straight(length=5)
    wg1 = gf.path.extrude(P1, X1)

    P3 = gf.path.straight(length=15, npoints=100)
    straight_transition = gf.path.extrude(P3, Xtrans)

    P4 = gf.path.euler(radius=25, angle=45, p=0.5, use_eff=False)
    wg_trans = gf.path.extrude(P4, Xtrans)
    wg_trans.show()

    # import gdsfactory as gf
    # # Create our first CrossSection
    # X1 = gf.CrossSection()
    # X1.add(width=1.2, offset=0, layer=2, name="wg", ports=("o1", "o2"))
    # X1.add(width=2.2, offset=0, layer=3, name="etch")
    # X1.add(width=1.1, offset=3, layer=1, name="wg2")

    # # Create the second CrossSection that we want to transition to
    # X2 = gf.CrossSection()
    # X2.add(width=1, offset=0, layer=2, name="wg", ports=("o1", "o2"))
    # X2.add(width=3.5, offset=0, layer=3, name="etch")
    # X2.add(width=3, offset=5, layer=1, name="wg2")

    # # To show the cross-sections, let's create two Paths and
    # # create Devices by extruding them
    # P1 = gf.path.straight(length=5)
    # P2 = gf.path.straight(length=5)
    # wg1 = gf.path.extrude(P1, X1)
    # wg2 = gf.path.extrude(P2, X2)

    # # Place both cross-section Devices and quickplot them
    # c = gf.Component()
    # wg1ref = c << wg1
    # wg2ref = c << wg2
    # wg2ref.movex(7.5)

    # # Create the transitional CrossSection
    # Xtrans = gf.path.transition(cross_section1=X1, cross_section2=X2, width_type="sine")
    # # Create a Path for the transitional CrossSection to follow
    # P3 = gf.path.straight(length=15, npoints=100)
    # # Use the transitional CrossSection to create a Component
    # straight_transition = gf.path.extrude(P3, Xtrans)

    # P4 = gf.path.euler(radius=25, angle=45, p=0.5, use_eff=False)
    # wg_trans = gf.path.extrude(P4, Xtrans)

    # c = gf.Component()
    # wg1_ref = c << wg1  # First cross-section Component
    # wg2_ref = c << wg2
    # wgt_ref = c << wg_trans

    # wgt_ref.connect("o1", wg1_ref.ports["o2"])
    # wg2_ref.connect("o1", wgt_ref.ports["o2"])



"""
info:

- default
- changed
- full
- info (derived properties)
- child: if any

Calculated/derived properties are stored in info

"""

import gdsfactory as gf


def test_args():
    c1 = gf.components.pad((150, 150))
    assert c1.info.full.size[0] == 150


if __name__ == "__main__":
    test_args()
    # assert c1.settings.size.full[0] == 150
    c1 = gf.components.pad((150, 150))
    c1.show()



from setuptools import find_packages, setup


def get_install_requires():
    with open("requirements.txt", "r") as f:
        return [line.strip() for line in f.readlines() if not line.startswith("-")]


def get_install_requires_dev():
    with open("requirements_dev.txt", "r") as f:
        return [line.strip() for line in f.readlines() if not line.startswith("-")]


def get_install_requires_full():
    with open("requirements_full.txt", "r") as f:
        return [line.strip() for line in f.readlines() if not line.startswith("-")]


setup(
    name="gdsfactory",
    url="https://github.com/gdsfactory/gdsfactory",
    version="4.0.13",
    author="gdsfactory community",
    scripts=["gdsfactory/gf.py"],
    description="python library to generate GDS layouts",
    packages=find_packages(),
    include_package_data=True,
    install_requires=get_install_requires(),
    python_requires=">=3.7",
    license="MIT",
    entry_points="""
        [console_scripts]
        gf=gdsfactory.gf:gf
    """,
    extras_require={
        "full": list(set(get_install_requires() + get_install_requires_full())),
        "basic": get_install_requires(),
        "dev": list(
            set(
                get_install_requires()
                + get_install_requires_dev()
                + get_install_requires_full()
            )
        ),
    },
)



"""# Components.
You can adapt some component functions from the `gdsfactory.components` module.
Each function there returns a Component object
"""


import gdsfactory as gf


def test_components_module() -> None:
    assert len(dir(gf.components)) > 1


if __name__ == "__main__":
    print(dir(gf.components))
    print(len(dir(gf.components)))



r"""
top view

.. code::

                          length
      <-|--------|--------------------------------->
        |        | length_section
        |<--------------------------->
       length_contact
        |<------>|
        |________|_____________________________
       /|        |____________________|        |
      / |viastack|                    |contact |
      \ | size   |____________________|        |
       \|________|____________________|________|
                                      |        |
                  cross_section_heater|        |
                                      |        |
                                      |        |
                                      |________|

cross_section

.. code::

                              |<------width------>|
      ____________             ___________________               ______________
     |            |           |     undoped Si    |             |              |
     |layer_heater|           |  intrinsic region |<----------->| layer_heater |
     |____________|           |___________________|             |______________|
                                                                 <------------>
                                                    heater_gap     heater_width

"""


from functools import partial

from gdsfactory.components.contact import contact_npp_m1
from gdsfactory.components.straight_heater_doped_rib import straight_heater_doped_rib
from gdsfactory.cross_section import strip_heater_doped

straight_heater_doped_strip = partial(
    straight_heater_doped_rib,
    cross_section_heater=strip_heater_doped,
    contact=contact_npp_m1,
)


if __name__ == "__main__":
    # c = straight_heater_doped_strip(length=100)
    # c = test_straight_heater_doped_strip_ports()
    c = straight_heater_doped_strip()
    c.show()



import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_s import bend_s
from gdsfactory.components.straight import straight as straight_function
from gdsfactory.cross_section import strip
from gdsfactory.types import ComponentFactory, CrossSectionFactory


@gf.cell
def coupler_asymmetric(
    bend: ComponentFactory = bend_s,
    straight: ComponentFactory = straight_function,
    gap: float = 0.234,
    dy: float = 5.0,
    dx: float = 10.0,
    cross_section: CrossSectionFactory = strip,
    **kwargs
) -> Component:
    """bend coupled to straight waveguide

    Args:
        bend:
        straight: straight library
        gap: um
        dy: port to port vertical spacing
        dx: bend length in x direction
        cross_section:
        **kwargs: cross_section settings

    .. code::

                        dx
                     |-----|
                      _____ o2
                     /         |
               _____/          |
         gap o1____________    |  dy
                            o3

    """
    x = cross_section(**kwargs)
    width = x.info["width"]
    bend_component = (
        bend(size=(dx, dy - gap - width), cross_section=cross_section, **kwargs)
        if callable(bend)
        else bend
    )
    wg = (
        straight(cross_section=cross_section, **kwargs)
        if callable(straight)
        else straight
    )

    w = bend_component.ports["o1"].width
    y = (w + gap) / 2

    c = Component()
    wg = wg.ref(position=(0, y), port_id="o1")
    bottom_bend = bend_component.ref(position=(0, -y), port_id="o1", v_mirror=True)

    c.add(wg)
    c.add(bottom_bend)

    # Using absorb here to have a flat cell and avoid
    # to have deeper hierarchy than needed
    c.absorb(wg)
    c.absorb(bottom_bend)

    port_width = 2 * w + gap
    c.add_port(name="o1", midpoint=[0, 0], width=port_width, orientation=180)
    c.add_port(port=bottom_bend.ports["o2"], name="o3")
    c.add_port(port=wg.ports["o2"], name="o2")

    return c


if __name__ == "__main__":
    c = coupler_asymmetric(gap=0.4, layer=(2, 0))
    c.show()



"""Lets for example customize the default gf.PDK

Fab A is mostly defined using Metal layers in GDS layer (30, 0)

The metal layer traces are 2um wide

"""

import gdsfactory as gf
from gdsfactory.cross_section import strip
from gdsfactory.difftest import difftest

WIDTH = 2
LAYER = (30, 0)

fab_a_metal = gf.partial(strip, width=WIDTH, layer=LAYER)
fab_a_metal.__name__ = "fab_a_metal"


straight = gf.partial(gf.components.straight, cross_section=fab_a_metal)
bend_euler = gf.partial(gf.components.bend_euler, cross_section=fab_a_metal)
mmi1x2 = gf.partial(
    gf.components.mmi1x2,
    cross_section=fab_a_metal,
    width=WIDTH,
    width_taper=WIDTH,
    width_mmi=3 * WIDTH,
)
mzi = gf.partial(gf.components.mzi, cross_section=fab_a_metal, splitter=mmi1x2)
gc = gf.partial(
    gf.components.grating_coupler_elliptical_te, layer=LAYER, wg_width=WIDTH
)


def test_waveguide():
    c = gf.components.straight(cross_section=fab_a_metal)
    difftest(c)


if __name__ == "__main__":

    # c = gf.components.straight(length=20, cross_section=fab_a_metal)
    c = mzi()
    wg_gc = gf.routing.add_fiber_array(
        component=c, grating_coupler=gc, cross_section=fab_a_metal
    )
    wg_gc.show()



import warnings

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler180
from gdsfactory.cross_section import strip
from gdsfactory.types import ComponentFactory, CrossSectionFactory


@gf.cell
def delay_snake2(
    length: float = 1600.0,
    length0: float = 0.0,
    n: int = 2,
    bend180: ComponentFactory = bend_euler180,
    cross_section: CrossSectionFactory = strip,
    **kwargs,
) -> Component:
    """Snake input facing west

    Snake output facing east

    This snakes can have a starting offset (length0)

    Args:
        length: total length
        length0: initial offset
        n: number of loops
        bend180
        cross_section: factory
        kwargs: cross_section settings

    .. code::

       | length0 | length1 |

                 >---------|
                           |  bend180.length
       |-------------------|
       |
       |------------------->

       |   delta_length    |


    """
    if n % 2:
        warnings.warn(f"rounding {n} to {n//2 *2}", stacklevel=3)
        n = n // 2 * 2

    bend180 = bend180(cross_section=cross_section, **kwargs)

    delta_length = (length - length0 - n * bend180.info.length) / (n + 1)
    length1 = delta_length - length0
    assert (
        length1 > 0
    ), "Snake is too short: either reduce length0, increase the total length,\
    or decrease n"

    s1 = gf.components.straight(length=length1, cross_section=cross_section, **kwargs)
    sd = gf.components.straight(
        cross_section=cross_section, length=delta_length, **kwargs
    )

    symbol_to_component = {
        "_": (s1, "o1", "o2"),
        "-": (sd, "o1", "o2"),
        ")": (bend180, "o2", "o1"),
        "(": (bend180, "o1", "o2"),
    }

    sequence = "_)" + n // 2 * "-(-)"
    sequence = sequence[:-1]
    c = gf.components.component_sequence(
        sequence=sequence, symbol_to_component=symbol_to_component
    )
    gf.port.auto_rename_ports(c)
    return c


def test_delay_snake2_length() -> Component:
    length = 500.0
    n = 6
    c = delay_snake2(n=n, length=length, layer=(2, 0))
    length_measured = (
        c.aliases[")1"].parent.info.length * n
        + c.aliases["-1"].parent.info.length * n
        + c.aliases["_1"].parent.info.length
    )
    assert np.isclose(
        length, length_measured
    ), f"length measured = {length_measured} != {length}"
    return c


if __name__ == "__main__":
    # c = test_delay_snake2_length()
    # c.show()
    c = delay_snake2(n=2, length=500, layer=(2, 0), length0=100)
    c.show()



"""Merge mask metadata with test labels to return test_metadata """
import warnings
from pathlib import Path
from typing import Any, Dict, List

from omegaconf import DictConfig

from gdsfactory.config import logger
from gdsfactory.types import PathType


def parse_csv_data(csv_labels_path: Path) -> List[List[str]]:
    """Returns CSV labels."""
    with open(csv_labels_path) as f:
        # Get all lines
        lines = [line.replace("\n", "") for line in f.readlines()]

        # Ignore labels for metrology structures
        lines = [line for line in lines if not line.startswith("METR_")]

        # Split lines in fields
        lines = [line.split(",") for line in lines]

        lines = [[s.strip() for s in splitted if s.strip()] for splitted in lines]

        # Remove empty lines
        lines = [line for line in lines if line]
    return lines


def get_cell_from_label(label: str) -> str:
    """get cell name from the label (cell_name is in parenthesis)"""
    cell_name = label.split("(")[1].split(")")[0]
    if cell_name.startswith("loopback"):
        cell_name = "_".join(cell_name.split("_")[1:])
    return cell_name


def merge_test_metadata(
    labels_path: PathType,
    mask_metadata: Dict[str, Any],
    labels_prefix: str = "opt",
) -> DictConfig:
    """Returns a test metadata dict config of labeled cells
    by merging GDS labels in CSV and YAML mask metadata

    Args:
        labels_path: for test labels in CSV
        mask_metadata: dict with test metadata
        labels_prefix: only select labels with a text prefix

    .. code::

        CSV labels  -------
                          |--> merge_test_metadata dict
                          |
        YAML metatada  ----


    """
    labels_path = Path(labels_path)

    if not labels_path.exists():
        raise FileNotFoundError(f"missing CSV labels {labels_path}")

    labels_list = parse_csv_data(labels_path)
    cells_metadata = mask_metadata.get("cells", {})

    test_metadata = DictConfig({})

    for label, x, y in labels_list:
        cell = get_cell_from_label(label)

        if cell in cells_metadata:
            test_metadata[cell] = cells_metadata[cell]
            test_metadata[cell].label = dict(x=float(x), y=float(y), text=label)
        else:
            logger.error(f"missing cell metadata for {cell}")
            warnings.warn(f"missing cell metadata for {cell}")

    return test_metadata


if __name__ == "__main__":
    from omegaconf import OmegaConf

    from gdsfactory import CONFIG

    labels_path = (
        CONFIG["samples_path"] / "mask_pack" / "build" / "mask" / "sample_mask.csv"
    )
    mask_metadata_path = labels_path.with_suffix(".yml")
    mask_metadata = OmegaConf.load(mask_metadata_path)
    d = merge_test_metadata(labels_path=labels_path, mask_metadata=mask_metadata)
    print(d)



from functools import partial

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.compass import compass
from gdsfactory.components.contact import contact_slab_npp_m3
from gdsfactory.tech import LAYER
from gdsfactory.types import ComponentFactory, Floats, Layers, Optional

pad_contact_slab_npp = partial(contact_slab_npp_m3, size=(80, 80))


@cell
def resistance_sheet(
    width: float = 10,
    layers: Layers = (LAYER.SLAB90, LAYER.NPP),
    layer_offsets: Floats = (0, 0.2),
    pad: ComponentFactory = pad_contact_slab_npp,
    pad_pitch: float = 100.0,
    ohms_per_square: Optional[float] = None,
) -> Component:
    """Sheet resistance.
    keeps connectivity for pads and first layer in layers

    Args:
        width:
        layers: for the middle part
        layer_offsets: from edge, positive: over, negative: inclusion
        pad: function to create a pad
        pad_pitch:
        ohms_per_square: optional sheet resistance to compute info.resistance
    """
    c = Component()

    pad = pad()
    length = pad_pitch - pad.info_child.size[0]

    pad1 = c << pad
    pad2 = c << pad
    r0 = c << compass(
        size=(length + layer_offsets[0], width + layer_offsets[0]), layer=layers[0]
    )

    for layer, offset in zip(layers[1:], layer_offsets[1:]):
        c << compass(size=(length + 2 * offset, width + 2 * offset), layer=layer)

    pad1.connect("e3", r0.ports["e1"])
    pad2.connect("e1", r0.ports["e3"])

    c.info_child.resistance = (
        ohms_per_square * width * length if ohms_per_square else None
    )

    c.add_port("pad1", port_type="vertical_dc", midpoint=pad1.center)
    c.add_port("pad2", port_type="vertical_dc", midpoint=pad2.center)
    return c


if __name__ == "__main__":
    c = resistance_sheet(width=40)
    c.show()



from pytest_regressions.data_regression import DataRegressionFixture

import gdsfactory as gf
from gdsfactory.component import Component


def test_get_bundle_west_to_north(
    data_regression: DataRegressionFixture, check: bool = True
) -> Component:

    lengths = {}

    c = gf.Component("test_get_bundle_west_to_north")
    pad = gf.partial(gf.components.pad, size=(10, 10))
    c = gf.Component()
    pad_south = gf.components.pad_array(orientation=270, spacing=(15.0, 0.0), pad=pad)
    pad_north = gf.components.pad_array(orientation=90, spacing=(15.0, 0.0), pad=pad)
    pl = c << pad_south
    pb = c << pad_north
    pl.rotate(90)
    pb.move((100, -100))

    pbports = pb.get_ports_list()
    ptports = pl.get_ports_list()

    c.add_ports(pbports, prefix="N")
    c.add_ports(ptports, prefix="S")

    routes = gf.routing.get_bundle(
        pbports,
        ptports,
        bend=gf.components.wire_corner,
    )
    for i, route in enumerate(routes):
        c.add(route.references)
        lengths[i] = route.length

    if check:
        data_regression.check(lengths)
    return c


def test_get_bundle_west_to_north2(
    data_regression: DataRegressionFixture, check: bool = True
) -> Component:

    lengths = {}
    c = gf.Component("test_get_bundle_west_to_north2")
    pbottom_facing_north = gf.port.port_array(
        midpoint=(0, 0), orientation=90, pitch=(30, 0)
    )
    ptop_facing_west = gf.port.port_array(
        midpoint=(100, 100), orientation=180, pitch=(0, -30)
    )

    routes = gf.routing.get_bundle(
        pbottom_facing_north,
        ptop_facing_west,
        bend=gf.components.wire_corner,
    )

    for i, route in enumerate(routes):
        c.add(route.references)
        lengths[i] = route.length

    if check:
        data_regression.check(lengths)
    return c


if __name__ == "__main__":
    c = test_get_bundle_west_to_north(None, check=False)
    # c = test_get_bundle_west_to_north2(None, check=False)
    c.show()



import collections
import multiprocessing
import pathlib
import time
from multiprocessing import Process
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from omegaconf import OmegaConf

from gdsfactory import components
from gdsfactory.config import CONFIG, logger
from gdsfactory.placer import (
    build_components,
    doe_exists,
    load_doe_component_names,
    save_doe,
)
from gdsfactory.sweep.read_sweep import get_settings_list
from gdsfactory.sweep.write_sweep import write_sweep_metadata
from gdsfactory.types import PathType

factory = {
    i: getattr(components, i)
    for i in dir(components)
    if not i.startswith("_") and callable(getattr(components, i))
}

n_cores = multiprocessing.cpu_count()


def separate_does_from_templates(dicts: Dict[str, Any]) -> Any:
    type_to_dict = {}

    does = {}
    for name, d in dicts.items():
        if "type" in d.keys():
            template_type = d.pop("type")
            if template_type not in type_to_dict:
                type_to_dict[template_type] = {}
            type_to_dict[template_type][name] = d
        else:
            does[name] = d

    return does, type_to_dict


def update_dicts_recurse(
    target_dict: Dict[
        str, Union[List[int], str, Dict[str, List[int]], Dict[str, str], bool]
    ],
    default_dict: Dict[str, Union[bool, Dict[str, Union[int, str]], int, str]],
) -> Dict[str, Any]:
    target_dict = target_dict.copy()
    default_dict = default_dict.copy()
    for k, v in default_dict.items():
        if k not in target_dict:
            target_dict[k] = v
        else:
            if isinstance(target_dict[k], (dict, collections.OrderedDict)):
                target_dict[k] = update_dicts_recurse(target_dict[k], default_dict[k])
    return target_dict


def save_doe_use_template(doe, doe_root_path=None) -> None:
    """Write a "content.txt" pointing to the DOE used as a template"""
    doe_name = doe["name"]
    doe_template = doe["doe_template"]
    doe_root_path = doe_root_path or CONFIG["cache_doe_directory"]
    doe_dir = doe_root_path / doe_name
    doe_dir.mkdir(exist_ok=True)
    content_file = doe_dir / "content.txt"

    with open(content_file, "w") as fw:
        fw.write(f"TEMPLATE: {doe_template}")


def write_sweep(
    doe,
    component_factory=factory,
    doe_root_path: Optional[PathType] = None,
    doe_metadata_path: Optional[PathType] = None,
    overwrite: bool = False,
    precision: float = 1e-9,
    **kwargs,
) -> None:
    doe_name = doe["name"]
    list_settings = doe["list_settings"]

    # Otherwise generate each component using the component library
    component_type = doe["component"]
    components = build_components(
        component_type, list_settings, component_factory=component_factory
    )

    component_names = [c.name for c in components]
    save_doe(doe_name, components, doe_root_path=doe_root_path, precision=precision)

    write_sweep_metadata(
        doe_name=doe["name"],
        cell_names=component_names,
        list_settings=doe["list_settings"],
        doe_settings=kwargs,
        doe_metadata_path=doe_metadata_path,
    )


def read_sweep(
    filepath: PathType, defaults: Optional[Dict[str, bool]] = None
) -> Tuple[Any, Any]:
    """Read DOE YAML file and returns a tuple of DOEs"""
    does = {}
    defaults = defaults or {"do_permutation": True, "settings": {}}
    data = OmegaConf.load(filepath)
    data = OmegaConf.to_container(data)
    mask = data.pop("mask")
    data.pop("vars", "")

    for doe_name, doe in data.items():
        for k in defaults:
            if k not in doe:
                doe[k] = defaults[k]

        does[doe_name] = doe
    return does, mask


def write_sweeps(
    filepath: PathType,
    component_factory: Dict[str, Callable] = factory,
    doe_root_path: PathType = CONFIG["cache_doe_directory"],
    doe_metadata_path: PathType = CONFIG["doe_directory"],
    n_cores: int = n_cores,
    overwrite: bool = False,
    precision: float = 1e-9,
    cache: bool = False,
) -> None:
    """Generates a sweep/DOEs of components specified in a yaml file
    allows for each DOE to have its own x and y spacing (more flexible than method1)
    similar to write_doe

    Args:
        filepath: for the does.yml
        component_factory:
        doe_root_path:
        doe_metadata_path:
        n_cores: number of cores
        overwrite:
        precision: for the GDS, defaults to 1nm
        cache: if True uses cache
    """
    doe_root_path = pathlib.Path(doe_root_path)
    doe_metadata_path = pathlib.Path(doe_metadata_path)

    doe_root_path.mkdir(parents=True, exist_ok=True)
    doe_metadata_path.mkdir(parents=True, exist_ok=True)

    dicts, mask_settings = read_sweep(filepath)
    does, templates_by_type = separate_does_from_templates(dicts)

    dict_templates = (
        templates_by_type["template"] if "template" in templates_by_type else {}
    )

    with_cache_default = mask_settings["cache"] if "cache" in mask_settings else cache

    list_args = []
    for doe_name, doe in does.items():
        doe["name"] = doe_name
        component = doe["component"]

        if component not in component_factory:
            raise ValueError(f"{component!r} not in {component_factory.keys()}")

        if "template" in doe:
            # The keyword template is used to enrich the dictionary from the template
            templates = doe["template"]
            if not isinstance(templates, list):
                templates = [templates]
            for template in templates:
                try:
                    doe = update_dicts_recurse(doe, dict_templates[template])
                except Exception:
                    print(template, "does not exist")
                    raise

        do_permutation = doe.pop("do_permutation")
        settings = doe["settings"]
        doe["list_settings"] = get_settings_list(do_permutation, **settings)

        list_args += [doe]

    does_running = []
    start_times = {}
    finish_times = {}
    doe_name_to_process = {}
    while list_args:
        while len(does_running) < n_cores:
            if not list_args:
                break
            doe = list_args.pop()
            doe_name = doe["name"]

            # Only launch a build process if we do not use the cache
            # Or if the DOE is not built

            list_settings = doe["list_settings"]

            with_cache = with_cache_default if "cache" not in doe else doe["cache"]

            _doe_exists = False

            if "doe_template" in doe:
                # this DOE points to another existing component
                _doe_exists = True
                logger.info(f"Using template - {doe_name!r}")
                save_doe_use_template(doe)

            elif with_cache:
                _doe_exists = doe_exists(
                    doe_name=doe_name,
                    list_settings=list_settings,
                    doe_root_path=doe_root_path,
                )
                if _doe_exists:
                    logger.info("Cached - {doe_name!r}")
                    if overwrite:
                        component_names = load_doe_component_names(doe_name)

                        write_sweep_metadata(
                            doe_name=doe["name"],
                            cell_names=component_names,
                            list_settings=doe["list_settings"],
                            doe_metadata_path=doe_metadata_path,
                        )

            if not _doe_exists:
                start_times[doe_name] = time.time()
                p = Process(
                    target=write_sweep,
                    args=(doe, component_factory),
                    kwargs={
                        "doe_root_path": doe_root_path,
                        "doe_metadata_path": doe_metadata_path,
                        "overwrite": overwrite,
                        "precision": precision,
                    },
                )
                doe_name_to_process[doe_name] = p
                does_running += [doe_name]
                try:
                    p.start()
                except Exception:
                    print(f"Issue starting process for {doe_name}")
                    print(type(component_factory))
                    raise

        to_rm = []
        for i, doe_name in enumerate(does_running):
            _p = doe_name_to_process[doe_name]
            if not _p.is_alive():
                to_rm += [i]
                finish_times[doe_name] = time.time()
                dt = finish_times[doe_name] - start_times[doe_name]
                line = "Done - {} ({:.1f}s)".format(doe_name, dt)
                logger.info(line)

        for i in to_rm[::-1]:
            does_running.pop(i)

        time.sleep(0.001)

    while does_running:
        to_rm = []
        for i, _doe_name in enumerate(does_running):
            _p = doe_name_to_process[_doe_name]
            if not _p.is_alive():
                to_rm += [i]
        for i in to_rm[::-1]:
            does_running.pop(i)

        time.sleep(0.05)


if __name__ == "__main__":
    filepath_sample = CONFIG["samples_path"] / "mask" / "does.yml"
    write_sweeps(filepath_sample, precision=2e-9)



"""autoplacer - placing GDS components with Klayout"""

from gdsfactory.autoplacer.auto_placer import AutoPlacer
from gdsfactory.autoplacer.chip_array import ChipArray
from gdsfactory.autoplacer.library import Library
from gdsfactory.autoplacer.yaml_placer import place_from_yaml

__all__ = ["AutoPlacer", "Library", "ChipArray", "place_from_yaml"]



"""Group references. Distribute them as you want

Lets distribute different references in the X direction

"""

if __name__ == "__main__":

    import gdsfactory as gf

    D = gf.Component()

    t1 = D << gf.components.text("1")
    t2 = D << gf.components.text("2")
    t3 = D << gf.components.text("3")
    t4 = D << gf.components.text("4")
    t5 = D << gf.components.text("5")
    t6 = D << gf.components.text("6")

    D.distribute(direction="x", spacing=3)

    gf.show(D)



"""Via cutback."""

from typing import Tuple

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.compass import compass
from gdsfactory.components.contact import contact_heater_m3
from gdsfactory.types import ComponentFactory, Float2


@gf.cell
def _via_iterable(
    via_spacing: float,
    wire_width: float,
    layer1: Tuple[int, int],
    layer2: Tuple[int, int],
    via_layer: Tuple[int, int],
    via_width: float,
) -> Component:
    """Via"""
    c = gf.Component()
    wire1 = c.add_ref(compass(size=(via_spacing, wire_width), layer=layer1))
    wire2 = c.add_ref(compass(size=(via_spacing, wire_width), layer=layer2))
    viac = c.add_ref(compass(size=(via_width, via_width), layer=via_layer))
    via1 = c.add_ref(compass(size=(via_width, via_width), layer=via_layer))
    wire1.connect(port="e3", destination=wire2.ports["e1"], overlap=wire_width)
    viac.connect(
        port="e1", destination=wire1.ports["e3"], overlap=(wire_width + via_width) / 2
    )
    via1.connect(
        port="e1", destination=wire2.ports["e3"], overlap=(wire_width + via_width) / 2
    )
    c.add_port(name="e1", port=wire1.ports["e1"], port_type="electrical")
    c.add_port(name="e3", port=wire2.ports["e3"], port_type="electrical")
    c.add_port(
        name="e4",
        midpoint=[(1 * wire_width) + wire_width / 2, -wire_width / 2],
        width=wire_width,
        orientation=-90,
        port_type="electrical",
    )
    c.add_port(
        name="e2",
        midpoint=[(1 * wire_width) + wire_width / 2, wire_width / 2],
        width=wire_width,
        orientation=90,
        port_type="electrical",
    )

    return c


@gf.cell
def via_cutback(
    num_vias: float = 100.0,
    wire_width: float = 10.0,
    via_width: float = 5.0,
    via_spacing: float = 40.0,
    min_pad_spacing: float = 0.0,
    pad: ComponentFactory = contact_heater_m3,
    pad_size: Float2 = (150, 150),
    layer1: Tuple[int, int] = gf.LAYER.HEATER,
    layer2: Tuple[int, int] = gf.LAYER.M1,
    via_layer: Tuple[int, int] = gf.LAYER.VIAC,
    wire_pad_inclusion: float = 12.0,
) -> Component:
    """Via cutback to extract via resistance

    adapted from phidl.geometry

    Args:
        num_vias: total requested vias needs to be even
        wire_width: width of wire
        via_width: width of via
        via_spacing: via_spacing
        pad_size: (width, height)
        min_pad_spacing
        pad_layer
        layer1: top wiring
        layer2: bottom wiring
        via_layer
        wire_pad_inclusion:

    """

    c = gf.Component()

    pad_component = pad(size=pad_size)
    pad1 = c.add_ref(pad_component)
    pad2 = c.add_ref(pad_component)

    nub = c.add_ref(pad(size=(3 * wire_width, wire_width)))
    head = c.add_ref(pad(size=(wire_width, wire_width)))
    nub.ymax = pad1.ymax - 5
    nub.xmin = pad1.xmax - wire_pad_inclusion
    head.connect(port="e1", destination=nub.ports["e3"])

    old_port = head.ports["e4"]
    count = 0
    width_via_iter = 2 * via_spacing - 2 * wire_width

    pad2.xmin = pad1.xmax + min_pad_spacing
    up = False
    down = True
    edge = True
    current_width = 3 * wire_width + wire_width  # width of nub and 1 overlap
    obj_old = head
    obj = head
    via_iterable = _via_iterable(
        via_spacing=via_spacing,
        wire_width=wire_width,
        layer1=layer1,
        layer2=layer2,
        via_layer=via_layer,
        via_width=via_width,
    )
    while (count + 2) <= num_vias:
        obj = c.add_ref(via_iterable)
        obj.connect(port="e1", destination=old_port, overlap=wire_width)
        old_port = obj.ports["e3"]
        edge = False
        if obj.ymax > pad1.ymax:
            obj.connect(port="e1", destination=obj_old.ports["e4"], overlap=wire_width)
            old_port = obj.ports["e4"]
            current_width += width_via_iter
            down = True
            up = False
            edge = True

        elif obj.ymin < pad1.ymin:
            obj.connect(port="e1", destination=obj_old.ports["e2"], overlap=wire_width)
            old_port = obj.ports["e2"]
            current_width += width_via_iter
            up = True
            down = False
            edge = True
        count = count + 2
        obj_old = obj

    if (
        current_width < min_pad_spacing
        and (min_pad_spacing - current_width) > 3 * wire_width
    ):
        tail = c.add_ref(
            pad(
                size=(min_pad_spacing - current_width + wire_width, wire_width),
            )
        )
    else:
        tail = c.add_ref(pad(size=(3 * wire_width, wire_width)))

    if up and not edge:
        tail.connect(port="e1", destination=obj.ports["e4"], overlap=wire_width)
    elif down and not edge:
        tail.connect(port="e1", destination=obj.ports["e2"], overlap=wire_width)
    else:
        tail.connect(port="e1", destination=obj.ports["e3"], overlap=wire_width)

    pad2.xmin = tail.xmax - wire_pad_inclusion
    return c


if __name__ == "__main__":
    c = via_cutback()
    c.show()
